from kooplearn._src.check_deps import check_torch_deps

check_torch_deps()

import torch  # noqa: E402

import kooplearn.nn.functional as F  # noqa: E402

__all__ = ["VAMPLoss", "DPLoss"]


class VAMPLoss():
    def __init__(self, schatten_norm: int = 2, center_covariances: bool = True):
        """Initializes the Variational Approach for learning Markov Processes (VAMP) loss by :footcite:t:`Wu2019`.

        Args:
            schatten_norm (int, optional): Computes the VAMP-p score with ``p = schatten_norm``. Defaults to 2.
            center_covariances (bool, optional): Use centered covariances to compute the VAMP score. Defaults to True.

        Raises:
            NotImplementedError: If ``schatten_norm`` is not 1 or 2.

        """
        if schatten_norm not in [1, 2]:
            raise NotImplementedError(f"Schatten norm {schatten_norm} not implemented")
        self.schatten_norm = schatten_norm
        self.center_covariances = center_covariances

    def __call__(self, X: torch.Tensor, Y: torch.Tensor):
        """Compute the VAMP loss function

        Args:
            X (torch.Tensor): Covariates for the initial time steps.
            Y (torch.Tensor): Covariates for the evolved time steps.
        """
        return -F.vamp_score(
            X,
            Y,
            schatten_norm=self.schatten_norm,
            center_covariances=self.center_covariances,
        )


class DPLoss():
    def __init__(
        self, relaxed: bool = True, metric_deformation: float = 1.0, center_covariances: bool = True
    ):
        """Initializes the (Relaxed) Deep Projection loss by :footcite:t:`Kostic2023DPNets`.

        Args:
            relaxed (bool, optional): Whether to use the relaxed (more numerically stable) or the full deep-projection loss. Defaults to True.
            metric_deformation (float, optional): Strength of the metric metric deformation loss: Defaults to 1.0.
            center_covariances (bool, optional): Use centered covariances to compute the VAMP score. Defaults to True.

        """
        self.relaxed = relaxed
        self.metric_deformation = metric_deformation
        self.center_covariances = center_covariances

    def __call__(self, X: torch.Tensor, Y: torch.Tensor):
        """Compute the Deep Projection loss function

        Args:
            X (torch.Tensor): Covariates for the initial time steps.
            Y (torch.Tensor): Covariates for the evolved time steps.
        """
        return -F.deepprojection_score(
            X,
            Y,
            relaxed=self.relaxed,
            metric_deformation=self.metric_deformation,
            center_covariances=self.center_covariances,
        )



import numpy as np

from kooplearn._src.utils import ShapeError
from kooplearn.abc import ContextWindow


def parse_observables(observables_dict, data: ContextWindow, data_fit: ContextWindow):
    if data.context_length != data_fit.context_length:
        raise ShapeError(
            f"The  context length ({data.context_length}) of the validation data does not match the context length of the training data ({data_fit.context_length})."
        )
    lookback_len = data.context_length - 1
    X_inference = data.lookback(lookback_len)
    X_fit, Y_fit = data_fit.lookback(lookback_len), data_fit.lookforward(lookback_len)

    if observables_dict is None:
        observables_dict = {"__state__": Y_fit}
    else:
        observables_dict["__state__"] = Y_fit

    parsed_obs = {}
    expected_shapes = {}
    for obs_name, obs in observables_dict.items():
        try:
            obs = np.asanyarray(obs)
        except Exception as e:
            raise ValueError(
                f"Observables must be convertible to a Numpy Array, while the following exception has been raised {e}"
            )

        if obs.dtype.kind != "f":
            raise TypeError(
                f"Observables should have floating-point values, whereas {obs_name} if of dtype {obs.dtype}"
            )
        if obs.shape[0] != observables_dict["__state__"].shape[0]:
            raise ShapeError(
                f"The observable {obs_name} was evaluated for {obs.shape[0]}, while the fitting data have {observables_dict['__state__'].shape[0]} examples."
            )
        # Reshape the observables to 2D arrays
        if obs.ndim == 1:
            obs = obs[:, None]

        # If the observables are multidimensional, flatten them and save the shape for the final reshape
        trailing_dims = obs.shape[1:]
        expected_shape = (X_inference.shape[0],) + trailing_dims
        if obs.ndim > 2:
            obs = obs.reshape(
                obs.shape[0], -1
            )  # Flatten out everything for proper broadcasting
        parsed_obs[obs_name] = obs
        expected_shapes[obs_name] = expected_shape

    return parsed_obs, expected_shapes, X_inference, X_fit


# !! Possibly to deprecate
def contexts_to_markov_train_states(
    contexts: ContextWindow,
    # lookback_len: int,
) -> np.ndarray:
    """TODO: Docstring for contexts_to_markov_IO_states.

    Args:
        contexts (np.ndarray): Array of contexts with shape ``(n_samples, context_len, *features_shape)``
        lookback_len (int): Length of the lookback window associated to the contexts.

        .. caution::

        If the lookforward window ``context_len - lookback_len - 1`` is not equal to 1, an error will be raised.

    Returns:
        tuple(np.ndarray, np.ndarray): TODO.
    """
    if not (contexts.shape[1] == contexts._lookback_len + 1):
        raise ShapeError(
            f"This function act on context windows with lookforward dimension == 1, while context_len = {contexts.shape[1]} and lookback_len = {contexts._lookback_len}, that is lookforward_len = {contexts.shape[1] - contexts._lookback_len} != 1"
        )

    _init = contexts.lookback(0).data
    _evolution = contexts.lookback(1).data
    return _init, _evolution


def contexts_to_markov_predict_states(
    contexts,
    # lookback_len: int,
) -> np.ndarray:
    """TODO: Docstring for contexts_to_markov_IO_states.

    Args:
        contexts (np.ndarray): Array of contexts with shape ``(n_samples, context_len, *features_shape)``
        lookback_len (int): Length of the lookback window associated to the contexts.

        .. caution::

        If the lookforward window ``context_len - lookback_len - 1`` is larger than 1, an error will be raised.

    Returns:
        tuple(np.ndarray, np.ndarray): TODO.
    """

    if contexts._lookback_len == contexts.shape[1]:
        X = contexts.data
        Y = None
    elif contexts._lookback_len + 1 == contexts.shape[1]:
        X = contexts.lookback(0).data
        Y = contexts.lookforward().data
    else:
        raise ShapeError(
            f"This function act on context windows with lookforward dimension == 0 or 1, while context_len = {contexts.shape[1]} and lookback_len = {contexts._lookback_len}, that is lookforward_len = {contexts.shape[1] - contexts._lookback_len} != 0 or 1"
        )
    return X, Y



from kooplearn.models.dict_of_fns import Linear, Nonlinear
from kooplearn.models.kernel import Kernel
from kooplearn.models.nystroem import NystroemKernel

try:
    from kooplearn._src.check_deps import check_torch_deps

    check_torch_deps()
    from kooplearn.models.ae.consistent import ConsistentAE
    from kooplearn.models.ae.dynamic import DynamicAE
except ImportError:
    pass



import numpy as np
from numpy.typing import ArrayLike

from kooplearn._src.check_deps import parse_backend
from kooplearn._src.utils import ShapeError
from kooplearn.abc import ContextWindowDataset


class TensorContextDataset(ContextWindowDataset):
    """
    Class for a collection of context windows with tensor features.
    """

    def __init__(self, data: ArrayLike, backend: str = "auto", **backend_kw):
        """
        Initializes the TensorContextDataset.

        Args:
            data (ArrayLike): A collection of context windows.
            backend (str, optional): Specifies the backend to be used (``'numpy'``, ``'torch'``). If set to ``'auto'``, will use the same backend of data. Defaults to ``'auto'``.
            **backend_kw (dict, optional): Keyword arguments to pass to the backend. For example, if ``'torch'``, it is possible to specify the device of the tensor.
        """
        # Backend selection
        torch, backend = parse_backend(backend)

        if backend == "numpy":
            if torch is not None:
                if torch.is_tensor(data):
                    data = data.detach().cpu().numpy()
                else:
                    data = np.asanyarray(data, **backend_kw)
            else:
                data = np.asanyarray(data, **backend_kw)
        elif backend == "torch":
            if torch is None:
                raise ImportError(
                    "You selected the 'torch' backend, but kooplearn wasn't able to import it."
                )
            else:
                if torch.is_tensor(data):
                    pass
                else:
                    data = torch.tensor(data, **backend_kw)
        elif backend == "auto":
            if torch is not None:
                if torch.is_tensor(data):
                    pass
                else:
                    data = np.asanyarray(data, **backend_kw)
            else:
                data = np.asanyarray(data, **backend_kw)

        # Attributes init
        if data.ndim < 3:
            raise ShapeError(
                f"Invalid shape {data.shape}. The data must have be at least three dimensional [batch_size, context_len, *features]."
            )
        self.data = data
        self.dtype = self.data.dtype
        self.shape = self.data.shape
        self.ndim = self.data.ndim
        self._context_length = self.shape[1]

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, idx):
        if np.issubdtype(type(idx), np.integer):
            return TensorContextDataset(self.data[idx][None, ...])
        elif isinstance(idx, slice):
            return TensorContextDataset(self.data[idx])
        else:
            try:
                return TensorContextDataset(self.data[idx])
            except Exception as e:
                raise (e)

    def slice(self, slice_obj):
        """
        Returns a slice of the context windows given a slice object.

        Args:
            slice_obj (slice): The python slice object.

        Returns:
            Slice of the context windows.
        """
        return self.data[:, slice_obj]


class TrajectoryContextDataset(TensorContextDataset):
    """
    Class for a collection of context windows with tensor features.
    """

    def __init__(
        self,
        trajectory: ArrayLike,
        context_length: int = 2,
        time_lag: int = 1,
        backend: str = "auto",
        **backend_kw,
    ):
        """
        Initializes the TrajectoryContextDataset. It takes as input a trajectory and returns a sequence of context windows.

        Args:
            trajectory (ArrayLike): A trajectory of shape ``(n_frames, *features_shape)``.
            context_length (int, optional): Length of the context window. Default to ``2``.
            time_lag (int, optional): Time lag, i.e. stride, between successive context windows. Default to ``1``.
            backend (str, optional): Specifies the backend to be used (``'numpy'``, ``'torch'``). If set to ``'auto'``, will use the same backend of the trajectory. Default to ``'auto'``.
            backend_kw (dict, optional): Keyword arguments to pass to the backend. For example, if ``'torch'``, it is possible to specify the device of the tensor.
        """
        if context_length < 1:
            raise ValueError(f"context_length must be >= 1, got {context_length}")

        if time_lag < 1:
            raise ValueError(f"time_lag must be >= 1, got {time_lag}")

        if trajectory.ndim < 1:
            raise ShapeError(
                f"Invalid trajectory shape {trajectory.shape}. The trajectory should be at least 1D."
            )
        elif trajectory.ndim == 1:
            trajectory = trajectory[:, None]  # Add a dummy dimension for 1D features
        else:
            pass

        # Backend selection

        torch, backend = parse_backend(backend)

        if backend == "numpy":
            if torch is not None:
                if torch.is_tensor(trajectory):
                    trajectory = trajectory.detach().cpu().numpy()
                    self.data, self.idx_map = _contexts_from_traj_np(
                        trajectory, context_length, time_lag
                    )
                else:
                    trajectory = np.asanyarray(trajectory, **backend_kw)
                    self.data, self.idx_map = _contexts_from_traj_np(
                        trajectory, context_length, time_lag
                    )
            else:
                trajectory = np.asanyarray(trajectory, **backend_kw)
                self.data, self.idx_map = _contexts_from_traj_np(
                    trajectory, context_length, time_lag
                )

        elif backend == "torch":
            if torch is None:
                raise ImportError(
                    "You selected the 'torch' backend, but kooplearn wasn't able to import it."
                )
            else:
                from kooplearn.nn.data import _contexts_from_traj_torch

                if not torch.is_tensor(trajectory):
                    trajectory = torch.tensor(trajectory, **backend_kw)
                self.data, self.idx_map = _contexts_from_traj_torch(
                    trajectory, context_length, time_lag
                )
        elif backend == "auto":
            if torch is not None:
                if torch.is_tensor(trajectory):
                    from kooplearn.nn.data import _contexts_from_traj_torch

                    self.data, self.idx_map = _contexts_from_traj_torch(
                        trajectory, context_length, time_lag
                    )
                else:
                    trajectory = np.asanyarray(trajectory, **backend_kw)
                    self.data, self.idx_map = _contexts_from_traj_np(
                        trajectory, context_length, time_lag
                    )
            else:
                trajectory = np.asanyarray(trajectory, **backend_kw)
                self.data, self.idx_map = _contexts_from_traj_np(
                    trajectory, context_length, time_lag
                )

        self.trajectory = trajectory
        self.time_lag = time_lag
        super().__init__(self.data)
        assert context_length == self.context_length


def _contexts_from_traj_np(trajectory, context_length, time_lag):
    window_shape = 1 + (context_length - 1) * time_lag
    if window_shape > trajectory.shape[0]:
        raise ValueError(
            f"Invalid combination of context_length={context_length} and time_lag={time_lag} for trajectory of "
            f"length {trajectory.shape[0]}. Try reducing context_length or time_lag."
        )

    data = np.lib.stride_tricks.sliding_window_view(trajectory, window_shape, axis=0)

    idx_map = np.lib.stride_tricks.sliding_window_view(
        np.arange(trajectory.shape[0], dtype=np.int_).reshape(-1, 1),
        window_shape,
        axis=0,
    )

    idx_map = np.moveaxis(idx_map, -1, 1)[:, ::time_lag, ...]
    data = np.moveaxis(data, -1, 1)[:, ::time_lag, ...]
    return data, TensorContextDataset(idx_map)


class MultiTrajectoryContextDataset(TrajectoryContextDataset):
    def __init__(self):
        raise NotImplementedError


def traj_to_contexts(
    trajectory: np.ndarray,
    context_window_len: int = 2,
    time_lag: int = 1,
    backend: str = "auto",
    **backend_kwargs,
):
    """
    Transforms a single trajectory to a sequence of context windows.

    Args:
        trajectory (np.ndarray): A trajectory of shape ``(n_frames, *features_shape)``.
        context_window_len (int, optional): Length of the context window. Default to ``2``.
        time_lag (int, optional): Time lag, i.e. stride, between successive context windows. Default to ``1``.
        backend (str, optional): Specifies the backend to be used (``'numpy'``, ``'torch'``). If set to ``'auto'``, will use the same backend of the trajectory. Default to ``'auto'``.
        backend_kw (dict, optional): Keyword arguments to pass to the backend. For example, if ``'torch'``, it is possible to specify the device of the tensor.

    Returns:
        TrajectoryContextDataset: A sequence of context windows.
    """
    return TrajectoryContextDataset(
        trajectory,
        context_length=context_window_len,
        time_lag=time_lag,
        backend=backend,
        **backend_kwargs,
    )



def check_torch_deps():
    try:
        import lightning
    except ImportError:
        raise ImportError(
            "To use kooplearn's deep learning models please reinstall it with the `torch` extra flag by typing `pip install kooplearn[torch]`."
        )


def parse_backend(backend: str):
    if backend not in ["auto", "numpy", "torch"]:
        raise ValueError(
            f"Invalid backend {backend}. Accepted values are 'auto', 'numpy', or 'torch'."
        )
    # Check if torch is available
    try:
        import torch
    except ImportError:
        torch = None
    return torch, backend


def check_dashboard_deps():
    try:
        import dash_iconify
        import dash_mantine_components
        import plotly
    except ImportError:
        raise ImportError(
            "To use kooplearn's dashboard please reinstall it with the `dashboard` extra flag by typing `pip install kooplearn[dashboard]`."
        )



import logging

import numpy as np
import torch

from kooplearn.data import TensorContextDataset

logger = logging.getLogger("kooplearn")


def collate_context_dataset(batch: list[TensorContextDataset]):
    concat_fn = torch.cat if torch.is_tensor(batch[0].data) else np.concatenate
    batched_data = torch.tensor(concat_fn([ctx.data for ctx in batch]))
    return TensorContextDataset(batched_data)


def _contexts_from_traj_torch(trajectory, context_length, time_lag):
    window_shape = 1 + (context_length - 1) * time_lag
    if window_shape > trajectory.shape[0]:
        raise ValueError(
            f"Invalid combination of context_length={context_length} and time_lag={time_lag} for trajectory of "
            f"length {trajectory.shape[0]}. Try reducing context_length or time_lag."
        )

    data = trajectory.unfold(0, window_shape, 1)
    idx_map = (torch.arange(len(trajectory)).reshape(-1, 1)).unfold(0, window_shape, 1)

    data = torch.movedim(data, -1, 1)[:, ::time_lag, ...]
    idx_map = torch.movedim(idx_map, -1, 1)[:, ::time_lag, ...]
    return data, TensorContextDataset(idx_map)



from kooplearn._src.check_deps import check_torch_deps

check_torch_deps()
from typing import Optional  # noqa: E402

import torch  # noqa: E402


def sqrtmh(A: torch.Tensor):
    # Credits to
    """Compute the square root of a Symmetric or Hermitian positive definite matrix or batch of matrices. Credits to  `https://github.com/pytorch/pytorch/issues/25481#issuecomment-1032789228 <https://github.com/pytorch/pytorch/issues/25481#issuecomment-1032789228>`_."""
    L, Q = torch.linalg.eigh(A)
    zero = torch.zeros((), device=L.device, dtype=L.dtype)
    threshold = L.max(-1).values * L.size(-1) * torch.finfo(L.dtype).eps
    L = L.where(L > threshold.unsqueeze(-1), zero)  # zero out small components
    return (Q * L.sqrt().unsqueeze(-2)) @ Q.mH


def covariance(X: torch.Tensor, Y: Optional[torch.Tensor] = None, center: bool = True):
    """Covariance matrix

    Args:
        X (torch.Tensor): Input covariates of shape ``(samples, features)``.
        Y (Optional[torch.Tensor], optional): Output covariates of shape ``(samples, features)`` Defaults to None.
        center (bool, optional): Whether to compute centered covariances. Defaults to True.

    Returns:
        torch.Tensor: Covariance matrix of shape ``(features, features)``. If ``Y is not None`` computes the cross-covariance between X and Y.
    """
    assert X.ndim == 2
    cov_norm = torch.rsqrt(torch.tensor(X.shape[0]))
    if Y is None:
        _X = cov_norm * X
        if center:
            _X = _X - _X.mean(dim=0, keepdim=True)
        return torch.mm(_X.T, _X)
    else:
        assert Y.ndim == 2
        _X = cov_norm * X
        _Y = cov_norm * Y
        if center:
            _X = _X - _X.mean(dim=0, keepdim=True)
            _Y = _Y - _Y.mean(dim=0, keepdim=True)
        return torch.mm(_X.T, _Y)


def vamp_score(X, Y, schatten_norm: int = 2, center_covariances: bool = True):
    """Variational Approach for learning Markov Processes (VAMP) score by :footcite:t:`Wu2019`.

    Args:
        X (torch.Tensor): Covariates for the initial time steps.
        Y (torch.Tensor): Covariates for the evolved time steps.
        schatten_norm (int, optional): Computes the VAMP-p score with ``p = schatten_norm``. Defaults to 2.
        center_covariances (bool, optional): Use centered covariances to compute the VAMP score. Defaults to True.

    Raises:
        NotImplementedError: If ``schatten_norm`` is not 1 or 2.

    """
    cov_X, cov_Y, cov_XY = (
        covariance(X, center=center_covariances),
        covariance(Y, center=center_covariances),
        covariance(X, Y, center=center_covariances),
    )
    if schatten_norm == 2:
        # Using least squares in place of pinv for numerical stability
        M_X = torch.linalg.lstsq(cov_X, cov_XY).solution
        M_Y = torch.linalg.lstsq(cov_Y, cov_XY.T).solution
        return torch.trace(M_X @ M_Y)
    elif schatten_norm == 1:
        sqrt_cov_X = sqrtmh(cov_X)
        sqrt_cov_Y = sqrtmh(cov_Y)
        M = torch.linalg.multi_dot(
            [
                torch.linalg.pinv(sqrt_cov_X, hermitian=True),
                cov_XY,
                torch.linalg.pinv(sqrt_cov_Y, hermitian=True),
            ]
        )
        return torch.linalg.matrix_norm(M, "nuc")
    else:
        raise NotImplementedError(f"Schatten norm {schatten_norm} not implemented")


def deepprojection_score(
    X,
    Y,
    relaxed: bool = True,
    metric_deformation: float = 1.0,
    center_covariances: bool = True,
):
    """Deep Projection score by :footcite:t:`Kostic2023DPNets`.

    Args:
        X (torch.Tensor): Covariates for the initial time steps.
        Y (torch.Tensor): Covariates for the evolved time steps.
        relaxed (bool, optional): Whether to use the relaxed (more numerically stable) or the full deep-projection loss. Defaults to True.
        metric_deformation (float, optional): Strength of the metric metric deformation loss: Defaults to 1.0.
        center_covariances (bool, optional): Use centered covariances to compute the VAMP score. Defaults to True.

    """
    cov_X, cov_Y, cov_XY = (
        covariance(X, center=center_covariances),
        covariance(Y, center=center_covariances),
        covariance(X, Y, center=center_covariances),
    )
    R_X = log_fro_metric_deformation_loss(cov_X)
    R_Y = log_fro_metric_deformation_loss(cov_Y)
    if relaxed:
        S = (torch.linalg.matrix_norm(cov_XY, ord="fro") ** 2) / (
            (
                torch.linalg.matrix_norm(cov_X, ord=2)
                * torch.linalg.matrix_norm(cov_Y, ord=2)
            )
        )
    else:
        M_X = torch.linalg.lstsq(cov_X, cov_XY).solution
        M_Y = torch.linalg.lstsq(cov_Y, cov_XY.T).solution
        S = torch.trace(M_X @ M_Y)
    return S - 0.5 * metric_deformation * (R_X + R_Y)


def log_fro_metric_deformation_loss(cov: torch.tensor):
    """Logarithmic + Frobenious metric deformation loss as used in :footcite:t:`Kostic2023DPNets`, defined as :math:`{{\\rm Tr}}(C^{2} - C -\ln(C))` .

    Args:
        cov (torch.tensor): A symmetric positive-definite matrix.

    Returns:
        torch.tensor: Loss function
    """
    eps = torch.finfo(cov.dtype).eps * cov.shape[0]
    vals_x = torch.linalg.eigvalsh(cov)
    vals_x = torch.where(vals_x > eps, vals_x, eps)
    loss = torch.mean(-torch.log(vals_x) + vals_x * (vals_x - 1.0))
    return loss



from __future__ import annotations

import abc
import os
from collections.abc import Iterable, Sequence
from typing import Optional

import numpy as np

from kooplearn._src.serialization import pickle_load, pickle_save


# Abstract base classes defining the interface to implement when extending kooplearn
class BaseModel(abc.ABC):
    @abc.abstractmethod
    def fit(self, *a, **kw):
        """Fit the model to the data. The signature of this function must be specified from the derived class. For example, in :class:`kooplearn.models.Nonlinear`, the signature is ``fit(self, data: np.ndarray)``, while in -TODO add AutoEncoder example."""
        pass

    @abc.abstractmethod
    def predict(
        self,
        data: ContextWindowDataset,
        t: int = 1,
        predict_observables: bool = True,
        reencode_every: int = 0,
    ):
        """
        Predicts the state or, if the system is stochastic, its expected value :math:`\mathbb{E}[X_t | X_0 = X]` after ``t`` instants given the initial conditions ``data.lookback(self.lookback_len)`` being the lookback slice of ``data``.
        If ``data.observables`` is not ``None``, returns the analogue quantity for the observable instead.

        Args:
            data (ContextWindowDataset): Dataset of context windows. The lookback window of ``data`` will be used as the initial condition, see the note above.
            t (int): Number of steps in the future to predict (returns the last one).
            predict_observables (bool): Return the prediction for the observables in ``data.observables``, if present. Default to ``True``.
            reencode_every (int): When ``t > 1``, periodically reencode the predictions as described in :footcite:t:`Fathi2023`. Only available when ``predict_observables = False``.

        Returns:
           The predicted (expected) state/observable at time :math:`t`. The result is composed of arrays with shape matching ``data.lookforward(self.lookback_len)`` or the contents of ``data.observables``. If ``predict_observables = True`` and ``data.observables != None``, the returned ``dict`` will contain the special key ``__state__`` containing the prediction for the state as well.
        """
        pass

    @abc.abstractmethod
    def eig(
        self,
        eval_left_on: Optional[ContextWindowDataset] = None,
        eval_right_on: Optional[ContextWindowDataset] = None,
    ):
        """
        Returns the eigenvalues of the Koopman/Transfer operator and optionally evaluates left and right eigenfunctions.

        Args:
            eval_left_on (ContextWindowDataset or None): Dataset of context windows on which the left eigenfunctions are evaluated.
            eval_right_on (ContextWindowDataset or None): Dataset of context windows on which the right eigenfunctions are evaluated.

        Returns:
            Eigenvalues of the Koopman/Transfer operator, shape ``(rank,)``. If ``eval_left_on`` or ``eval_right_on`` are not ``None``, returns the left/right eigenfunctions evaluated at ``eval_left_on`` / ``eval_right_on`` : shape ``(n_samples, rank)``.
        """
        pass

    @abc.abstractmethod
    def save(self, path: os.PathLike):
        pass

    @classmethod
    @abc.abstractmethod
    def load(path: os.PathLike):
        pass

    @abc.abstractmethod
    def modes(
        self,
        data: ContextWindowDataset,
        predict_observables: bool = True,
    ):
        """
        Computes the mode decomposition of arbitrary observables of the Koopman/Transfer operator at the states defined by ``data``.

        Informally, if :math:`(\\lambda_i, \\xi_i, \\psi_i)_{i = 1}^{r}` are eigentriplets of the Koopman/Transfer operator, for any observable :math:`f` the i-th mode of :math:`f` at :math:`x` is defined as: :math:`\\lambda_i \\langle \\xi_i, f \\rangle \\psi_i(x)`. See :footcite:t:`Kostic2022` for more details.

        Args:
            data (TensorContextDataset): Dataset of context windows. The lookback window of ``data`` will be used as the initial condition, see the note above.
            predict_observables (bool): Return the prediction for the observables in ``data.observables``, if present. Default to ``True``.
        Returns:
            (modes, eigenvalues): Modes and corresponding eigenvalues of the system at the states defined by ``data``. The result is composed of arrays with shape matching ``data.lookforward(self.lookback_len)`` or the contents of ``data.observables``. If ``predict_observables = True`` and ``data.observables != None``, the returned ``dict`` will contain the special key ``__state__`` containing the modes for the state as well.
        """
        pass

    @property
    @abc.abstractmethod
    def is_fitted(self) -> bool:
        """Check if the model is fitted.

        Returns:
            Returns ``True`` if the model is fitted, ``False`` otherwise.
        """
        pass

    @property
    @abc.abstractmethod
    def lookback_len(self) -> int:
        """Length of the lookback window associated to the contexts. Upon fitting, the dimension of the lookforward window will be inferred from the context window length and this attribute. Moreover, shape checks against this attribute will be performed on the data passed to :func:`fit`, :func:`predict`, :func:`eig` and :func:`modes`.

        Returns:
            Length of the lookback window associated to the contexts.
        """
        pass


class FeatureMap(abc.ABC):
    """Abstract base class for feature maps. The :func:`__call__` method must accept a batch of data points of shape ``(n_samples, *features_shape)`` and return a batch of features of shape ``(n_samples, out_features)``.

    .. caution::

        As described in :ref:`kooplearn's data paradigm <kooplearn_data_paradigm>`, the inputs passed to the methods of :class:`BaseModel` will be in the form of batches of context windows. :class:`FeatureMap`, instances are a notable departure from this paradigm, and we expect a feature map to be called on batches of data points, not context windows. The context windows are parsed internally by :class:`BaseModel` and passed to the feature map as batches of appropriate data points.

        This behaviour is designed to facilitate the reuse of feature maps across different models, and possibly even outside of kooplearn.
    """

    @abc.abstractmethod
    def __call__(self, data: np.ndarray) -> np.ndarray:
        pass


class TrainableFeatureMap(FeatureMap):
    @abc.abstractmethod
    def fit(self, *a, **kw) -> None:
        """Fit the feature map to the data."""
        pass

    @property
    @abc.abstractmethod
    def is_fitted(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def lookback_len(self) -> int:
        """Length of the lookback window associated to the contexts. Upon fitting, the dimension of the lookforward window will be inferred from the context window length.

        Returns:
            Length of the lookback window associated to the contexts.
        """
        pass


class ContextWindow(Sequence):
    """
    Class for a single context window, i.e. the :ref:`kooplearn's data paradigm <kooplearn_data_paradigm>`.
    """

    def __init__(self, window: Sequence):
        """
        Initializes the context window.

        Args:
            window (Sequence): A sequence of data points.
        """
        self.data = window
        self._context_length = len(window)

    @property
    def context_length(self):
        """
        Returns the length of the context window.
        """
        return self._context_length

    def slice(self, slice_obj):
        """
        Returns a slice of the context window given a slice object.

        Args:
            slice_obj (slice): The python slice function.

        Returns:
            Slice of the context window.
        """
        return self.data[slice_obj]

    def __len__(self):
        """
        Returns the length of the context window.
        """
        return 1  # Single context window

    def lookback(self, lookback_length: int, slide_by: int = 0):
        """
        Returns the lookback window of the context window.

        Args:
            lookback_length (int):  Length of the lookback window.
            slide_by (int, optional): Number of slides along the context window. Default to ``0``.

        Returns:
            Lookback window of the context window.
        """
        self._check_lb_len(lookback_length)
        max_slide = self._context_length - lookback_length
        if slide_by > max_slide:
            raise ValueError(
                f"Invalid slide_by = {slide_by} for lookback_length = {lookback_length} and Context of length = {self._context_length}. It should be 0 <= slide_by <= context_length - lookback_length"
            )

        lb_window = self.slice(slice(slide_by, lookback_length + slide_by))
        return lb_window

    def lookforward(self, lookback_length: int):
        """
        Returns the lookforward window of the context window.

        Args:
            lookback_length (int): Length of the lookback window.

        Returns:
            Lookforward window of the context window.
        """
        self._check_lb_len(lookback_length)
        lf_window = self.slice(slice(lookback_length, None))
        return lf_window

    def _check_lb_len(self, lookback_length: int):
        """
        Checks the validity of the lookback length.

        Args:
            lookback_length (int): Length of the lookback window.
        """
        if (lookback_length > self.context_length) or (lookback_length < 1):
            raise ValueError(
                f"Invalid lookback_length = {lookback_length} for ContextWindow of length = {self.context_length}. It should be 1 <= lookback_length <= context_length."
            )

    def __repr__(self):
        """
        Returns a string representation of the context window.
        """
        return f"ContextWindow <context_length={self.context_length}, data={self.data.__str__()}>"

    def __getitem__(self, idx):
        """
        Returns the context window's point at the given index.

        Args:
            idx (int): The index of the point to be returned.

        Returns:
            The point at the given index.
        """
        return self.data[idx]

    def save(self, path: os.PathLike):
        """
        Saves the current context window to the given path.

        Args:
            path (path-like): Save the context window to path.
        """
        pickle_save(self, path)

    @classmethod
    def load(cls, filename):
        """
        Loads the context window from the given filename.

        Args:
            filename: Load the context window from the given filename.

        Returns:
            The context window object.
        """
        return pickle_load(cls, filename)


class ContextWindowDataset(ContextWindow):
    """
    Class for a collection of :obj:`ContextWindow` objects.
    """

    def __init__(self, dataset: Iterable[Sequence]):
        """
        Initializes the context window dataset.

        Args:
            dataset (Iterable[Sequence]): A sequence of :obj:`ContextWindow` objects.
        """
        data = []
        context_lengths = []
        for ctx in dataset:
            if isinstance(ctx, Sequence):
                if isinstance(ctx, ContextWindow):
                    context_lengths.append(ctx.context_length)
                else:
                    context_lengths.append(len(ctx))
                data.append(ctx)
            else:
                raise TypeError(
                    f"{self.__class__.__name__} should be initialized with an iterable of ContextWindow objects, while an object of {type(ctx)} was provided."
                )
        context_lengths = set(context_lengths)
        if len(context_lengths) != 1:
            raise ValueError(
                "The context windows in the dataset should all have the same length."
            )
        self.data = data
        self._context_length = context_lengths.pop()

    def __len__(self):
        """
        Returns the number of context windows in the dataset.
        """
        return len(self.data)

    def __iter__(self):
        """
        Returns an iterator over the context windows in the dataset
        """
        return iter([ContextWindow(ctx) for ctx in self.data])

    def __getitem__(self, idx):
        """
        Returns the context window at the given index.

        Args:
            idx: The index of the context window to be returned.

        Returns:
            The context window at the given index.
        """
        return ContextWindow(self.data[idx])

    def __repr__(self):
        """
        Returns a string representation of the context window dataset.
        """
        return f"{self.__class__.__name__} <item_count={len(self)}, context_length={self.context_length}, data={self.data.__str__()}>"

    def slice(self, slice_obj):
        """
        Returns a slice of the context window dataset given a slice object.

        Args:
            slice_obj (slice): The python slice object.

        Returns:
            Slice of the context window dataset.
        """
        return [ctx[slice_obj] for ctx in self.data]



from typing import Optional, Union

import torch

from kooplearn._src.utils import ShapeError
from kooplearn.data import TensorContextDataset


def encode_contexts(contexts_batch: TensorContextDataset, encoder: torch.nn.Module):
    # Caution: this method is designed only for internal calling.
    context_len = contexts_batch.context_length
    batch_size = len(contexts_batch)
    trail_dims = contexts_batch.shape[2:]

    encoded_contexts = encoder(
        contexts_batch.data.view(batch_size * context_len, *trail_dims)
    )
    encoded_contexts = encoded_contexts.view(batch_size, context_len, -1)
    return TensorContextDataset(encoded_contexts)


def decode_contexts(
    encoded_contexts_batch: TensorContextDataset, decoder: torch.nn.Module
):
    # Caution: this method is designed only for internal calling.
    context_len = encoded_contexts_batch.context_length
    batch_size = len(encoded_contexts_batch)
    assert (
        len(encoded_contexts_batch.shape[2:]) == 1
    ), "The decoder input must be a 1-dimensional tensor (i.e. a vector)."

    decoded_contexts = decoder(
        encoded_contexts_batch.data.view(batch_size * context_len, -1)
    )
    trail_dims = decoded_contexts.shape[1:]
    decoded_contexts = decoded_contexts.view(batch_size, context_len, *trail_dims)
    return TensorContextDataset(decoded_contexts)


def evolve_contexts(
    encoded_contexts_batch: TensorContextDataset,
    lookback_len: int,
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
    backward_operator: Optional[Union[torch.nn.Parameter, torch.Tensor]] = None,
):
    # Caution: this method is designed only for internal calling.
    context_len = encoded_contexts_batch.context_length

    X_init = torch.squeeze(
        encoded_contexts_batch.slice(slice(lookback_len - 1, lookback_len))
    )  # Initial condition
    if X_init.ndim != 2:
        raise ShapeError(
            f"The encoder network must return a 1D vector for each snapshot, while a shape {X_init.shape[1:]} tensor was received."
        )
    evolved_contexts_batch = torch.zeros_like(encoded_contexts_batch.data)

    for ctx_idx in range(context_len):  # Not efficient but working
        exp = ctx_idx - lookback_len + 1
        if exp < 0:
            if backward_operator is None:
                raise ValueError(
                    "No backward operator provided, but lookback length greater than 1."
                )
            else:
                pwd_operator = torch.matrix_power(backward_operator, -exp)
                Z = torch.mm(pwd_operator, X_init.T).T
        else:
            pwd_operator = torch.matrix_power(forward_operator, exp)
            Z = torch.mm(pwd_operator, X_init.T).T
        evolved_contexts_batch[:, ctx_idx, ...] = Z
    return TensorContextDataset(
        evolved_contexts_batch
    )  # [batch_size, context_len, latent_dim]


def evolve_forward(
    batch: TensorContextDataset,
    lookback_len: int,
    t: int,
    encoder: torch.nn.Module,
    decoder: torch.nn.Module,
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
):
    init_contexts = batch.slice(
        slice(lookback_len - 1, lookback_len)
    )  # Should be of shape [batch_size, 1, *trail_dims]
    init_contexts = init_contexts.view(len(batch), *batch.shape[2:])
    encoded_init_contexts = encoder(init_contexts)
    _raw_encoded_inits = encoded_init_contexts.data.view(len(init_contexts), -1)
    pwd_operator = torch.matrix_power(forward_operator, t)
    _raw_evolved_inits = torch.mm(pwd_operator, _raw_encoded_inits.T).T
    _raw_decoded_preds = decoder(_raw_evolved_inits)
    return TensorContextDataset(
        _raw_decoded_preds.view(len(batch), 1, *batch.shape[2:])
    )


def evolve_batch(
    batch: TensorContextDataset,
    t: int,
    encoder: torch.nn.Module,
    decoder: torch.nn.Module,
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
):
    Z = encoder(batch)
    assert (
        len(Z.shape[1:]) == 1
    ), "The encoder output must be a 1-dimensional tensor (i.e. a vector)."
    evolution_operator = torch.matrix_power(forward_operator, t)
    Z_evolved = torch.mm(evolution_operator, Z.T).T
    batch_evolved = decoder(Z_evolved)
    batch_evolved = batch_evolved.reshape(batch.shape)
    assert batch.shape == batch_evolved.shape
    return TensorContextDataset(batch_evolved)


def consistency_loss(
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
    backward_operator: Union[torch.nn.Parameter, torch.Tensor],
):
    assert (
        forward_operator.shape == backward_operator.shape
    ), "The forward and backward operators must have the same shape."
    terms = []
    dim = forward_operator.shape[0]
    for k in range(1, dim + 1):
        fb = torch.mm(forward_operator[:k, :], backward_operator[:, :k]) - torch.eye(
            k, device=forward_operator.device
        )
        bf = torch.mm(backward_operator[:k, :], forward_operator[:, :k]) - torch.eye(
            k, device=forward_operator.device
        )
        terms.append(
            torch.linalg.matrix_norm(fb) ** 2 + torch.linalg.matrix_norm(bf) ** 2
        )
    return 0.5 * torch.mean(torch.stack(terms))



from pathlib import Path
from shutil import rmtree

import numpy as np
import pytest
from scipy.stats import special_ortho_group
from sklearn.gaussian_process.kernels import RBF, DotProduct, Matern

from kooplearn.data import traj_to_contexts
from kooplearn.datasets.stochastic import LinearModel
from kooplearn.models import Kernel

TRUE_RANK = 5
DIM = 20
NUM_SAMPLES = 50


def make_linear_system():
    eigs = 9 * np.logspace(-3, -1, TRUE_RANK)
    # print("Eigenvalues:")
    # for ev in eigs:
    #     print(f"{ev:.1e} \t", end='')
    eigs = np.concatenate([eigs, np.zeros(DIM - TRUE_RANK)])
    Q = special_ortho_group(DIM, 0).rvs(1)
    A = np.linalg.multi_dot([Q, np.diag(eigs), Q.T])

    # Consistency-check
    assert np.allclose(np.sort(np.linalg.eigvalsh(A)), np.sort(eigs))
    return LinearModel(A, noise=1e-5, rng_seed=0)


@pytest.mark.parametrize("kernel", [DotProduct(), RBF(), Matern()])
@pytest.mark.parametrize("reduced_rank", [True, False])
@pytest.mark.parametrize("rank", [TRUE_RANK, TRUE_RANK - 2, TRUE_RANK + 10])
@pytest.mark.parametrize("solver", ["full", "arnoldi", "wrong"])
@pytest.mark.parametrize("tikhonov_reg", [None, 0.0, 1e-5])
@pytest.mark.parametrize("observables", [None, {"zeroes": np.zeros, "ones": np.ones}])
@pytest.mark.parametrize("predict_observables", [True, False])
@pytest.mark.parametrize("lookback_len", [1, 2, 3])
def test_Kernel_fit_predict_eig_modes_save_load(
    kernel,
    reduced_rank,
    rank,
    solver,
    tikhonov_reg,
    observables,
    predict_observables,
    lookback_len,
):
    dataset = make_linear_system()
    _Z = dataset.sample(np.zeros(DIM), NUM_SAMPLES)
    data = traj_to_contexts(_Z, lookback_len + 1)
    if solver not in ["full", "arnoldi"]:
        with pytest.raises(ValueError):
            model = Kernel(
                kernel=kernel,
                reduced_rank=reduced_rank,
                rank=rank,
                tikhonov_reg=tikhonov_reg,
                svd_solver=solver,
            )
    else:
        model = Kernel(
            kernel=kernel,
            reduced_rank=reduced_rank,
            rank=rank,
            tikhonov_reg=tikhonov_reg,
            svd_solver=solver,
        )

        assert model.is_fitted is False
        model.fit(data)
        assert model.is_fitted is True

        if (observables is None) or (predict_observables is False):
            data.observables = observables
            X_pred = model.predict(data, predict_observables=predict_observables)
            assert X_pred.shape == data.lookforward(model.lookback_len).shape
            modes, _ = model.modes(data, predict_observables=predict_observables)
            assert (
                modes.shape
                == (model.rank,) + data.lookforward(model.lookback_len).shape
            )
        else:
            obs_shape = (len(data), 1, 2, 3, 4)
            data.observables = {
                k: v(obs_shape, dtype=np.float_) for k, v in observables.items()
            }

            X_pred = model.predict(data, predict_observables=predict_observables)
            assert "__state__" in X_pred.keys()
            for k, v in X_pred.items():
                if k == "__state__":
                    assert v.shape == data.lookforward(model.lookback_len).shape
                else:
                    assert v.shape == obs_shape

            modes, _ = model.modes(data, predict_observables=predict_observables)
            assert "__state__" in modes.keys()
            for k, v in modes.items():
                if k == "__state__":
                    assert (
                        v.shape
                        == (model.rank,) + data.lookforward(model.lookback_len).shape
                    )
                else:
                    assert v.shape == (model.rank,) + obs_shape

        vals, lv, rv = model.eig(eval_left_on=data, eval_right_on=data)
        assert vals.shape[0] <= rank
        assert vals.ndim == 1
        tmp_path = Path(__file__).parent / f"tmp/model.bin"
        model.save(tmp_path)
        restored_model = Kernel.load(tmp_path)

        assert np.allclose(model.kernel_X, restored_model.kernel_X)
        assert np.allclose(model.kernel_Y, restored_model.kernel_Y)
        assert np.allclose(model.kernel_YX, restored_model.kernel_YX)
        assert np.allclose(model.data_fit, restored_model.data_fit)
        assert np.allclose(model.lookback_len, model.lookback_len)
        rmtree(Path(__file__).parent / "tmp/")



import abc
from typing import NamedTuple

import numpy as np
from tqdm import tqdm


class LinalgDecomposition(NamedTuple):
    values: np.ndarray
    x: np.ndarray
    functions: np.ndarray


# General utility classes
class DataGenerator(abc.ABC):
    @abc.abstractmethod
    def sample(self, initial_condition: np.ndarray, T: int = 1):
        pass


class DiscreteTimeDynamics(DataGenerator):
    def sample(
        self, initial_condition: np.ndarray, T: int = 1, show_progress: bool = False
    ):
        """Base Class to implement a discrete-time dataset

        Args:
            initial_condition (np.ndarray): Initial condition
            T (int, optional): Number of steps of the evolution. Defaults to 1.

        Returns:
            np.ndarray: An array of shape ``(T + 1, *initial_condition.shape)``
        """
        initial_condition = np.asarray(initial_condition)
        if initial_condition.ndim == 0:
            initial_condition = initial_condition[None]

        memory = np.zeros((T + 1,) + initial_condition.shape)
        memory[0] = initial_condition
        if show_progress:
            _iter = tqdm(range(T), desc="Generating data", unit="step", leave=False)
        else:
            _iter = range(T)

        for t in _iter:
            memory[t + 1] = self._step(memory[t])
        return memory

    @abc.abstractmethod
    def _step(self, X: np.ndarray):
        pass
